#ifndef HPMODEL_H
#define HPMODEL_H

#include <vector>
#include <string>
#include <stdexcept>
#include <iostream>

enum DIRECTION {FORWARD, UP, DOWN, LEFT, RIGHT};
enum ABSOLUTE_DIRECTION {NORTH, EAST, SOUTH, WEST, ABS_UP, ABS_DOWN};
enum ROTATION {NO_ROTATION, ROTATE_90, ROTATE_180, ROTATE_270};

class Point {
public:
    int x, y, z;
    bool hydrophobic;
    Point(const int x, const int y, const int z, const bool hydrophobic);

    void print() const;
    bool coords_match(Point other) const;
};

class HPModel {
private:
    std::string sequence;
    std::vector<Point> acids;
    ABSOLUTE_DIRECTION currentDirection;
    ABSOLUTE_DIRECTION previousDirection;
    ROTATION rotation;
    Point getHead();
    static std::string directionToString(DIRECTION direction);
    bool isInsertValid(Point toInsert);
    bool isCurrentAcidHydrophobic();
    static ABSOLUTE_DIRECTION getPlaneDirectionChange(ABSOLUTE_DIRECTION absolute_direction, DIRECTION relative_direction);
    ABSOLUTE_DIRECTION getUpDownDirectionAndUpdatePrevious(DIRECTION relative_direction);
    void updateCurrentDirection(DIRECTION direction);
    void addPoint(DIRECTION direction, bool isHydrophobic);
    void rotate(int amount);
    DIRECTION getRotatedRelativeDirection(DIRECTION relative_direction);
    static ABSOLUTE_DIRECTION invertDirection(ABSOLUTE_DIRECTION direction);

public:
    std::vector<Point> getPoints();
    void fold(DIRECTION direction);
    HPModel(const std::string& sequence);
    void print() const;
    std::string pointsToJson() const;
};

#endif
