//
// Created by Adam Wahab on 04/09/2024.
//

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <filesystem>
#include "gtest/gtest.h"
#include "../../src/HPModel.h"
#include <stdexcept>

TEST(basic_check, test_eq) {
    EXPECT_EQ(1,1);
}

TEST(basic_check, test_fold_quantity) {
    HPModel model = HPModel("hp");
    model.fold(FORWARD);
    EXPECT_THROW(model.fold(FORWARD), std::invalid_argument);

    model = HPModel("pppp");
    model.fold(FORWARD);
    model.fold(FORWARD);
    model.fold(FORWARD);
    EXPECT_THROW(model.fold(FORWARD), std::invalid_argument);

    EXPECT_THROW(model = HPModel(""),  std::invalid_argument);
    EXPECT_THROW(model = HPModel("p"),  std::invalid_argument);
}

TEST(basic_check, test_circular) {
    HPModel model = HPModel("hpphp");
    model.fold(RIGHT);
    model.fold(RIGHT);
    model.fold(RIGHT);
    EXPECT_THROW(model.fold(RIGHT), std::invalid_argument);
}

TEST(basic_check, test_planar) {
    HPModel model = HPModel("ppppp");
    model.fold(FORWARD);
    model.fold(RIGHT);
    model.fold(FORWARD);
    model.fold(LEFT);
    auto points = model.getPoints();

    // Initial point
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Forward
    Point p1 = points[1];
    Point p1Match = Point(0, 0, 1, false);

    // Right
    Point p2 = points[2];
    Point p2Match = Point(1, 0, 1, false);

    // Forward
    Point p3 = points[3];
    Point p3Match = Point(2, 0, 1, false);

    // Left
    Point p4 = points[4];
    Point p4Match = Point(2, 0, 2, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

TEST(basic_check, test_up_loop_valid) {
    HPModel model = HPModel("pppp");
    model.fold(UP);
    model.fold(UP);
    model.fold(UP);

    auto points = model.getPoints();

    // Initial Point
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Up
    Point p1 = points[1];
    Point p1Match = Point(0, 1, 0, false);

    // Up
    Point p2 = points[2];
    Point p2Match = Point(0, 1, -1, false);

    // Up
    Point p3 = points[3];
    Point p3Match = Point(0, 0, -1, false);

    model.print();

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
}

TEST(basic_check, test_up_relative_valid) {
    HPModel model = HPModel("ppppp");
    model.fold(UP);
    model.fold(RIGHT);
    model.fold(LEFT);
    model.fold(LEFT);

    auto points = model.getPoints();

    // Initial Point
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Up
    Point p1 = points[1];
    Point p1Match = Point(0, 1, 0, false);

    // Right
    Point p2 = points[2];
    Point p2Match = Point(1, 1, 0, false);

    // Left
    Point p3 = points[3];
    Point p3Match = Point(1, 2, 0, false);

    // Left
    Point p4 = points[4];
    Point p4Match = Point(0, 2, 0, false);

    model.print();

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

TEST(basic_check, test_up_loop_invalid) {
    HPModel model = HPModel("ppppp");
    model.fold(UP);
    model.fold(UP);
    model.fold(UP);
    EXPECT_THROW(model.fold(UP), std::invalid_argument);
    model.print();
}


TEST(basic_check, test_up_down) {
    HPModel model = HPModel("pppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(FORWARD);
    model.fold(DOWN);
    model.fold(LEFT);
    auto points = model.getPoints();

    // Initial Point
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Right
    Point p1 = points[1];
    Point p1Match = Point(1, 0, 0, false);

    // Up
    Point p2 = points[2];
    Point p2Match = Point(1, 1, 0, false);

    // Forward
    Point p3 = points[3];
    Point p3Match = Point(1, 2, 0, false);

    // Down
    Point p4 = points[4];
    Point p4Match = Point(2, 2, 0, false);

    // Left
    Point p5 = points[5];
    Point p5Match = Point(2, 2, 1, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
    EXPECT_TRUE(p5.coords_match(p5Match));
}

TEST(basic_check, big_test) {
    HPModel model = HPModel("pppppppp");
    model.fold(DOWN);
    model.fold(DOWN);
    model.fold(RIGHT);
    model.fold(DOWN);
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(LEFT);

    auto points = model.getPoints();

    model.print();

    // Initial point
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Down
    Point p1 = points[1];
    Point p1Match = Point(0, -1, 0, false);

    // Down
    Point p2 = points[2];
    Point p2Match = Point(0, -1, -1, false);

    // Right
    Point p3 = points[3];
    Point p3Match = Point(1, -1, -1, false);

    // Down
    Point p4 = points[4];
    Point p4Match = Point(1, 0, -1, false);

    // Right
    Point p5 = points[5];
    Point p5Match = Point(1, 0, 0, false);

    // Up
    Point p6 = points[6];
    Point p6Match = Point(2, 0, 0, false);

    // Left
    Point p7 = points[7];
    Point p7Match = Point(2, 1, 0, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
    EXPECT_TRUE(p5.coords_match(p5Match));
    EXPECT_TRUE(p6.coords_match(p6Match));
    EXPECT_TRUE(p7.coords_match(p7Match));
}

TEST(basic_check, right_up) {
    HPModel model = HPModel("ppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(UP);
    model.fold(RIGHT);

    auto points = model.getPoints();

    // Initial
    Point p0 = points[0];
    Point p0Match = Point(0, 0, 0, false);

    // Right
    Point p1 = points[1];
    Point p1Match = Point(1, 0, 0, false);

    // Up
    Point p2 = points[2];
    Point p2Match = Point(1, 1, 0, false);

    // Up
    Point p3 = points[3];
    Point p3Match = Point(0, 1, 0, false);


    // Right
    Point p4 = points[4];
    Point p4Match = Point(0, 1, -1, false);

    model.print();

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

// JSON tests
std::string readJsonFile(const std::string& filename) {
    std::ifstream file(filename);

    if (!file.is_open()) {
        std::cerr << "Unable to open file: " << filename << std::endl;
        return "";
    }

    std::stringstream buffer;
    buffer << file.rdbuf();
    return buffer.str();
}

TEST(basic_check, json_points_encodes_correctly_with_no_hydrophobics) {
    HPModel model = HPModel("ppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(UP);
    model.fold(RIGHT);

    auto modelJsonData = model.pointsToJson();

    std::string filename =  std::string(TEST_FILES_DIR) + "/points_no_hydrophobes.json";
    std::string premadeJsonData = readJsonFile(filename);
    EXPECT_EQ(modelJsonData, premadeJsonData);
}

// TODO
TEST(basic_check, json_points_encodes_correctly_with_hydrophobics) {

}